
%   eye_gaze_estimation: performs eye gaze estimation using the MAPS
%   algorithm for both Isomap an LLE as described in [MAPS].
%
%   BEGIN COPYRIGHT NOTICE
%
%   MAPS code -- (c) 2013-2014 Hamid Dadkhahi
%
%    This code is provided as is, with no guarantees except that 
%    bugs are almost surely present.  Published reports of research 
%    using this code (or a modified version) should cite the 
%    article that describes the algorithm: 
%
%      [MAPS] H. Dadkhahi, M. F. Duarte, Masking Strategies for Image
%      Manifolds, 2014.
%
%    Comments and bug reports are welcome.  Email to hdadkhahi@ecs.umass.edu. 
%    I would also appreciate hearing about how you used this code, 
%    improvements that you have made to it, or translations into other
%    languages.    
%
%    You are free to modify, extend or distribute this code, as long 
%    as this copyright notice is included whole and unchanged. 
%
%    END COPYRIGHT NOTICE

clear all;

% loading data matrix X
load X_eyedata.mat;

% dimensionality of the dataset X
d = size(X,1);
% number of data points in dataset X
n = size(X,2);

% size of the neighborhood used in either Isomap or LLE algorithms
K = 12;
K_LLE = K;
K_Isomap = K;

% masking size m
m = 50:50:300;
m_high = max(m);
m_low = min(m);

% vectors of eye gaze estimation error for different masking/embedding
% schemes
error_original = zeros(1,length(m));
error_PCA = zeros(1,length(m));
error_SPCA = zeros(1,length(m));
error_PCoA = zeros(1,length(m));
error_LLE = zeros(1,length(m));
error_Isomap = zeros(1,length(m));
error_random = zeros(1,length(m));

% PCoA
omega_PCoA = generate_omega_PCoA(X,m_high);

% LLE Masking Algorithm
omega_LLE = LLE_graph_algo(X,K_LLE,m_high);

% Isomap Masking Algorithm
distances_ambient = L2_distance(X, X, 1);
[Y_isomap,~,E_isomap,distances_geodesic] = Isomap2(distances_ambient,'k',K_Isomap);
[x_ind,y_ind] = find( triu(double(E_isomap) - eye(size(E_isomap,1))) );
A = ( X(:,x_ind) - X(:,y_ind) ).^2;
normA = sum(A,1);
A = A./repmat(normA,d,1);
omega_Isomap = greedy_algo(A,m_high);
clear A normA x_ind y_ind E_isomap;

% number of iterations for the random mask
n_iterations = 10;

for i = 1:length(m)
    
    % Isomap    
    Y_Isomap = X(omega_Isomap(1:m(i)),:);
    
    % LLE
    Y_LLE = X(omega_LLE(1:m(i)),:);
    
    % PCoA
    Y_PCoA = X(omega_PCoA(1:m(i)),:);
    
    % SPCA
    [cards,vars,Z]= sparsePCA(X',m(i));
    omega_SPCA = find( abs(Z) > 0.01 );
    Y_SPCA = X(omega_SPCA(1:m(i)),:);
    
    temp_random = zeros(1,n_iterations);
    % Random
    for j = 1 : n_iterations
        omega_random = randperm(d,m(i));
        Y_random = X(omega_random,:);
        temp_random(j) = mapping_function(Y_random,labels,K);
    end
        
    % PCA 
    [U,S,V] = svds(X,m(i));
    Y_PCA = U'*X;
        
    error_original(i) = mapping_function(X,labels,K);
    error_LLE(i) = mapping_function(Y_LLE,labels,K);
    error_random(i) = mean(temp_random);
    error_Isomap(i) = mapping_function(Y_Isomap,labels,K);
    error_PCA(i) = mapping_function(Y_PCA,labels,K);
    error_PCoA(i) = mapping_function(Y_PCoA,labels,K);
    error_SPCA(i) = mapping_function(Y_SPCA,labels,K);
    
end

figure, plot(m,error_original,'Color','b','LineStyle','--','Marker','x','LineWidth',2)
hold on, plot(m,error_PCA,'Color','c','LineStyle','-.','Marker','p','LineWidth',2)
hold on, plot(m,error_SPCA,'Color',[1 0.5 0],'LineStyle','-','Marker','d','LineWidth',2)
hold on, plot(m,error_PCoA,'Color','g','LineStyle','-','Marker','none','LineWidth',2)
hold on, plot(m,error_LLE,'Color','r','LineStyle','-','Marker','+','LineWidth',2)
hold on, plot(m,error_Isomap,'Color','k','LineStyle','-','Marker','v','LineWidth',2)
hold on, plot(m,error_random,'Color','m','LineStyle','-','Marker','o','LineWidth',2)
legend('Full Data','PCA','SPCA','PCoA','MAPS-LLE','MAPS-Isomap','Random','Location','NorthEast');
axis tight
axisfortex('','Embedding Dim./Masking Size m','Average Gaze Estimation Error')
